function [R,profile] = profchol(format,R,profile,epsilon)
% ----------------------------------------------------------------------
% In-place profile (skyline) Cholesky decomposition of a real
% symmetric positive (semi-)definite n x n matrix A, represented in
% possibly-packed form by matrix R.  A is stored in R by columns and
% only its upper triangle is accessed. A's profile shape is specified
% by the n-element row vector profile: in column i, only elements from
% that specified by profile(i) down to the diagonal are accessed.  R can
% be in one of three different formats, profile being optional for the
% first two:
%
% 1. R is a dense n x n matrix, profile(i) is the index of the first
% nonzero entry in column i and only elements R(profile(i):i,i) in the
% upper triangle of R are accessed.
%
% 2. A is a band matrix with above-diagonal bandwidth b (1<=b<=n). R
% is a b x n matrix containing the above-diagonal and diagonal entries
% of A, stored by columns. Only elements R(profile(i):b,i) are
% accessed, representing upper triangle elements A(i+profile(i)-b:i,i)
% of A. R(b,:) is the diagonal of A.
%
% 3. A is stored in fully packed by-columns form as a long column
% vector R.  profile(i+1) gives the index of diagonal element i in R,
% the remainder of column i being stored immediately before
% this. I.e. R(profile(i)+1:profile(i+1)) contains the nonzero entries
% A(i+profile(i)-profile(i+1)+1:i,i) of A. R(profile) is the diagonal
% of A.
%
% If epsilon>=0 is specified, any diagonal elements of the output
% matrix that would be less than epsilon or imaginary are increased to
% epsilon. Otherwise imaginary entries are errors.
% ----------------------------------------------------------------------
% For comparison, here's the dense version of columnwise Cholesky:
%
% for i=1:n
%   for j=1:i-1
%     R(j,i) = (R(j,i)-sum(R(1:j-1,j).*R(1:j-1,i))) / R(j,j);
%   end
%   R(i,i) = sqrt(R(i,i)-sum(R(1:i-1,i).^2));
% end
% ----------------------------------------------------------------------
% NOTE: For Matlab 6.5 and higher, just-in-time compilation makes
% for() loops much faster than building index matrices. for()-izing
% speeds up the inner loops below by a factor of about 7 in post-6.5
% Matlab, but slows them down by a factor of about 10 in pre-6.5
% Matlab and Octave. Set the comments below according to your version.

[b,n] = size(R);
if nargin<3, profile=[]; end
if nargin<4, epsilon=-1; end
eps2 = max(0,epsilon)^2;

if format == 1		% Unpacked n x n matrix.

   % Find profile if it wasn't given.
   if isempty(profile)
      profile = zeros(1,n);
      for i=1:n
	 nz = find([R(1:i,i);1]);
	 profile(i) = nz(1); 
      end
   end

   % Columnwise profile Cholesky.
   for i=1:n
      p = profile(i);
      for j=p:i-1
	 if 0
	    % Pre-6.5 / Octave version
	    K = max(p,profile(j)) : j-1;
	    R(j,i) = (R(j,i)-sum(R(K,j).*R(K,i))) / R(j,j);
	 else
	    % Post-6.5 version
	    t = R(j,i);
	    for k = max(p,profile(j)) : j-1
	       t = t - R(k,j)*R(k,i);
	    end
	    R(j,i) = t / R(j,j);
	 end
      end
      a = R(i,i)-sum(R(p:i-1,i).^2);
      if a>=eps2,	 R(i,i) = sqrt(a); 
      elseif epsilon>=0, R(i,i) = epsilon; 
      else error(['profchol: matrix nonpositive at column ',int2str(i)]);
      end
   end

elseif format == 2	% packed band matrix, bandwidth b

   % Find profile if it wasn't given.
   if isempty(profile)
      profile = zeros(1,n);
      for i=1:n 
	 nz = find([R(:,i);1]); 
	 profile(i) = nz(1);
      end
   end

   % Columnwise profile Cholesky. R(profile(k):b,k) represents elements
   % A(k+profile(k)-b:k,k) of unpacked matrix.
   for i=1:n
      p = profile(i);
      for j = p+i-b : i-1
	 if 0
	    % Pre-6.5 / Octave version
	    K = max(p+i-j,profile(j)) : b-1;
	    R(j-i+b,i) = (R(j-i+b,i)-sum(R(K,j).*R(K+(j-i),i))) / R(b,j);
	 else
	    % Post-6.5 version
	    ji = j-i;
	    t = R(ji+b,i);
	    for k = max(p-ji,profile(j)) : b-1
	       t = t - R(k,j)*R(k+ji,i);
	    end
	    R(ji+b,i) = t / R(b,j);
	 end
      end
      a = R(b,i)-sum(R(p:b-1,i).^2);
      if a>=eps2,	 R(b,i) = sqrt(a); 
      elseif epsilon>=0, R(b,i) = epsilon; 
      else error(['profchol: matrix nonpositive at column ',int2str(i)]);
      end
   end

elseif format == 3	% fully packed profile matrix

   if isempty(profile)
      error('profchol: for packed matrix, you must specify the profile');
   end
   n = size(profile,2)-1;

   % Columnwise profile Cholesky. Column i is in R(profile(i-1)+1:profile(i)),
   % representing elements A(profile(i-1)+1-profile(i)+i,i) of unpacked
   % matrix, and similarly for j.

   for i=1:n
      di = profile(i+1);
      bi = di-i;
      i0 = profile(i)-bi+1;
      for j = i0 : i-1
	 dj = profile(j+1);
	 bj = dj-j;
	 j0 = profile(j)-bj+1;
	 if 0
	    % Pre-6.5 / Octave version
	    K = max(i0,j0):j-1;
	    R(bi+j) = (R(bi+j)-sum(R(bi+K).*R(bj+K))) / R(dj);
	 else
	    % Post-6.5 version
	    t = R(bi+j);
	    for k = max(i0,j0) : j-1
	       t = t - R(bi+k)*R(bj+k);
	    end
	    R(bi+j) = t / R(dj);
	 end
      end
      a = R(di)-sum(R(bi+i0:di-1).^2);
      if a>=eps2,	 R(di) = sqrt(a); 
      elseif epsilon>=0, R(di) = epsilon; 
      else error(['profchol: matrix nonpositive at column ',int2str(i)]);
      end
   end
   % R'
else
   error('profchol: unkonwn matrix format');
end
